#region usings
using System;
using System.ComponentModel.Composition;
using System.Collections.Generic;
using System.Drawing;
using System.Windows.Forms;

using VVVV.PluginInterfaces.V1;
using VVVV.PluginInterfaces.V2;
using VVVV.PluginInterfaces.V2.Graph;

using VVVV.Core.Logging;

// Below for handling XML Strings
using System.Xml;
using System.Xml.Linq;
using System.IO;

#endregion usings

namespace VVVV.Nodes
{
	#region PluginInfo
	[PluginInfo(Name = "MakeSomeSpace", 
				Category = "VVVV",
				Help = "Allows insertion of empty space into a patch below a selection of nodes via CTRL-ALT-SHIFT + LeftDrag",
				Author = "mxmi",
				AutoEvaluate = false)]
	#endregion PluginInfo
	public class VVVVMakeSomeSpaceNode : IPluginEvaluate
	{
		#region fields & pins
		[Import()]
		public ILogger FLogger;
		
		[Import()]
		IHDEHost FHDEHost;
	
		private bool FDisposed;
		private INode2[] FSelectedNodes;
		private bool FMoveNodes = false;
		private Point FCursorStart;
		private Cursor FCursor;
		
		#endregion fields & pins

		#region constructor/destructor
		[ImportingConstructor]
		public VVVVMakeSomeSpaceNode(IHDEHost host, ILogger logger)
		{
			FLogger = logger; 

			FHDEHost = host;
			FHDEHost.NodeSelectionChanged += this.NodeSelectionChangedCB;
			FHDEHost.MouseDown += this.MouseDownCB;
			FHDEHost.MouseUp += this.MouseUpCB;
		}
		~VVVVMakeSomeSpaceNode()
		{
			Dispose(false);
		}
		public void Dispose()
		{
			Dispose(true);
		}
		protected void Dispose(bool disposing)
		{
			// Check to see if Dispose has already been called.
			if(!FDisposed)
			{
				if(disposing)
				{
					// Dispose managed resources.
					FHDEHost.NodeSelectionChanged -= this.NodeSelectionChangedCB;
					FHDEHost.MouseDown -= this.MouseDownCB;
					FHDEHost.MouseUp -= this.MouseUpCB;
				}
				// Release unmanaged resources. If disposing is false,
				// only the following code is executed.
			}
			FDisposed = true;
		}
		#endregion
		
		private void NodeSelectionChangedCB(Object sender, NodeSelectionEventArgs args)
		{
			if (FMoveNodes == false)
			{
				FSelectedNodes = args.Nodes;
			}
		}
		private void MouseDownCB(Object sender, VVVV.PluginInterfaces.V2.MouseEventArgs args)
		{
			if (args.Button == Mouse_Buttons.Left && args.ModifierKey == (Modifier_Keys.Control | Modifier_Keys.Alt | Modifier_Keys.Shift) && FSelectedNodes != null && FSelectedNodes.Length != 0)
			{
				FCursorStart = Cursor.Position;
				FMoveNodes = true;
				
				FCursor = Cursor.Current;
				Cursor.Current = Cursors.HSplit;
				
				// restore original selection so we know whats happening?
				FHDEHost.SelectNodes(FSelectedNodes);
			}
		}
		private void MouseUpCB(Object sender, VVVV.PluginInterfaces.V2.MouseEventArgs args)
		{
			if (FMoveNodes == true)
			{
				if (args.Button == Mouse_Buttons.Left && args.ModifierKey == (Modifier_Keys.Control | Modifier_Keys.Alt | Modifier_Keys.Shift))
				{
					int offsetPx = (Cursor.Position.Y - FCursorStart.Y);
					MoveNodesAndLinks(offsetPx);
				}
				FMoveNodes = false;
				
				Cursor.Current = FCursor;
				
				// restore original selection so we can chain movements
				FHDEHost.SelectNodes(FSelectedNodes);
			}
		}		
		private (INode2 Node, int Position) GetLowestNodeWithPosition(INode2[] nodes)
		{
			INode2 lowestNode = null;
			
			foreach (INode2 node in nodes)
			{
				if (lowestNode == null)
				{
					lowestNode = node;
				}
				else if (node.GetBounds(BoundsType.Node).Bottom > lowestNode.GetBounds(BoundsType.Node).Bottom)
				{
					lowestNode = node;
				}
			}
			
			return (Node: lowestNode, Position: lowestNode.GetBounds(BoundsType.Node).Bottom);
		}
		private List<INode2> GetLowerSiblings(INode2 node)
		{
			List<INode2> siblings = new List<INode2>();

			int refPosition = node.GetBounds(BoundsType.Node).Bottom;
			
			foreach (INode2 n in node.Parent)
			{
				if (n.GetBounds(BoundsType.Node).Bottom > refPosition)
				{
					siblings.Add(n);
				}
			}
			return siblings;
		}
		private static Stream GenerateStreamFromString(string s)
		{
		    var stream = new MemoryStream();
		    var writer = new StreamWriter(stream);
		    writer.Write(s);
		    writer.Flush();
		    stream.Position = 0;
		    return stream;
		}
		private XElement GetXElementFromString(string xmlString)
		{
			using (Stream stream = GenerateStreamFromString(xmlString))
			{
				XmlReaderSettings settings = new XmlReaderSettings();
				settings.ProhibitDtd = false;
				XmlReader reader = XmlReader.Create(stream, settings);
				XDocument doc = XDocument.Load(reader);
				return doc.Root;
			}
		}
		private void MoveNodesAndLinks(int offsetPx)
		{
			// get lowest node and its position
			(INode2 Node, int Position) lowest = GetLowestNodeWithPosition(FSelectedNodes);
			
			// get all nodes in LNode.parent that are lower than LNode -> NodesToMove
			List<INode2> lowerSiblings = GetLowerSiblings(lowest.Node);
			
			if (lowerSiblings.Count > 0)
			{
				// Move all NodesToMove down by x px
				// HDE host select all lowerSibling Nodes
				FHDEHost.SelectNodes(lowerSiblings.ToArray());
				
				// HDE Host getXML Snippet
				// make XElement from string
				XElement selection;
				selection = GetXElementFromString(FHDEHost.GetXMLSnippetFromSelection());
				
				// create xml for setting new position
				string xmlString = "<PATCH>";
				
				// iterate over <NODE>s
				foreach (XElement nodeElement in selection.Elements("NODE"))
				{
					xmlString += "<NODE id=\"" + nodeElement.Attribute("id").Value + "\">";
					
					// iterate over <BOUNDS>s
					foreach (XElement boundsElement in nodeElement.Elements("BOUNDS"))
					{
						if (boundsElement.Attribute("type").Value == "Node" || boundsElement.Attribute("type").Value == "Box")
						{
							xmlString += "<BOUNDS type=\"" + boundsElement.Attribute("type").Value + "\"";
							
							// add offest
							int position;
							Int32.TryParse(boundsElement.Attribute("top").Value, out position);
							position += 15 * offsetPx;
							
							xmlString += " top=\"" + position + "\"/>";
						}
					}
					xmlString += "</NODE>";
				}
				// iterate over <LINK>s which - luckily - are part of the selection xml string as well
				foreach (XElement linkElement in selection.Elements("LINK"))
				{
					//hope this changes the link accordingly
					xmlString += String.Format("<LINK srcnodeid=\"{0}\" srcpinname=\"{1}\" dstnodeid=\"{2}\" dstpinname=\"{3}\">", linkElement.Attribute("srcnodeid").Value, linkElement.Attribute("srcpinname").Value, linkElement.Attribute("dstnodeid").Value, linkElement.Attribute("dstpinname").Value);
					
					// iterate over <LINKPOINT>s
					foreach (XElement linkpointElement in linkElement.Elements("LINKPOINT"))
					{
						// calculate new position
						int position;
						Int32.TryParse(linkpointElement.Attribute("y").Value, out position);
						
						if (position > lowest.Position)
						{
							position += 15 * offsetPx;
						}
						xmlString += String.Format("<LINKPOINT x=\"{0}\" y=\"{1}\"/>", linkpointElement.Attribute("x").Value, position);
					}
					xmlString += "</LINK>";
				}
				
				xmlString += "</PATCH>";
				
				// and sent snippet
				FHDEHost.SendXMLSnippet(lowest.Node.Parent.NodeInfo.Filename, xmlString, true);
			}
		}
		public void Evaluate(int SpreadMax)
		{
		}
	}
}
