#region usings
using System;
using System.Collections.Generic;
using System.ComponentModel.Composition;

using System.Windows.Forms;
using System.Runtime.InteropServices;

using VVVV.PluginInterfaces.V1;
using VVVV.PluginInterfaces.V2;
using VVVV.PluginInterfaces.V2.Graph;

using VVVV.Core.Logging;
#endregion usings

namespace VVVV.Nodes
{
	#region PluginInfo
	[PluginInfo(Name = "WindowMode", 
				Category = "VVVV", 
				Help = "Automatically changes Patch, Editor and Renderer Windows' appearance", 
				Author = "mxmi",
				AutoEvaluate = false)]
	#endregion PluginInfo
	
	public class WindowModeNode : IPluginEvaluate, IDisposable
	{
		#region externs & constants
		//Sets window attributes
		[DllImport("USER32.DLL")]
		static extern int SetWindowLong(IntPtr hWnd, int nIndex, int dwNewLong);
		
		//Gets window attributes
		[DllImport("USER32.DLL")]
		static extern int GetWindowLong(IntPtr hWnd, int nIndex);
		
		[DllImport("USER32.DLL")]
		static extern IntPtr SetWindowPos(IntPtr hWnd, int hWndInsertAfter, int x, int Y, int cx, int cy, int wFlags);		

		//Menu editing 
		[DllImport("user32.dll")]
    	static extern int GetSystemMenu(IntPtr hWnd, bool bRevert);
		
		[DllImport("user32.dll")]
        static extern int GetMenuItemCount(IntPtr hMenu);
		
		[DllImport("user32.dll")]
		static extern bool DrawMenuBar(IntPtr hWnd);
		
		[DllImport("user32.dll")]
		static extern bool RemoveMenu(IntPtr hMenu, uint uPosition, uint uFlags);
		
		//assorted constants needed
		private const uint MF_BYCOMMAND  = 0x00000000;
    	private const uint MF_BYPOSITION = 0x00000400;
    	private const uint SC_CLOSE      = 0xF060;
		
		public static int GWL_STYLE = -16;
		public static int GWL_EXSTYLE = -20;
		
		public static int WS_CHILD = 0x40000000; //child window
		public static int WS_BORDER = 0x00800000; //window with border
		public static int WS_CAPTION = 0x00C00000; //window with a title bar 
		public static int WS_SYSMENU = 0x00080000; //window menu  
		public static int WS_VISIBLE = 0x10000000; //window visible 
		
		public static int WS_MAXIMIZEBOX = 0x00010000; //maximize box 
		public static int WS_MINIMIZEBOX = 0x00020000; //minimize box
		
		public static int WS_SIZEBOX = 0x00040000; //resizeable border 
		
		
		public static int WS_EX_APPWINDOW = 0x00040000; //top-level window
		public static int WS_EX_WINDOWEDGE = 0x00000100;
		
		
		public static int SWP_NOMOVE = 0x0002; //dont move  
		public static int SWP_NOSIZE = 0x0001; //dont resize  
		public static int SWP_NOZORDER = 0x0004; //dont rearrange z  
		public static int SWP_FRAMECHANGED = 0x0020; //indicate frame change  
		
		
		#endregion

		#region fields & pins
 		[Import()]
        public ILogger FLogger;
		
		[Import()]
	   	#pragma warning restore
		IHDEHost FHDEHost;
		
		private bool FDisposed;
		#endregion fields & pins
		
		#region constructor & destructor
		[ImportingConstructor]
		public WindowModeNode(IHDEHost host, ILogger logger)
		{
			FLogger = logger; 

			FHDEHost = host;
			FHDEHost.WindowAdded += WindowAddedCB;
			FHDEHost.AfterComponentModeChange += ComponentModeChangedCB;
		}
		
		~WindowModeNode()
		{
			Dispose(false);
		}
		
		public void Dispose()
		{
			Dispose(true);
		}
		
		protected void Dispose(bool disposing)
		{
			// Check to see if Dispose has already been called.
			if(!FDisposed)
			{
				if(disposing)
				{
					// Dispose managed resources.
					FHDEHost.WindowAdded -= WindowAddedCB;
					FHDEHost.AfterComponentModeChange -= ComponentModeChangedCB;
				}
				// Release unmanaged resources. If disposing is false,
				// only the following code is executed.
			}
			FDisposed = true;
		}
		#endregion
		
		private void WindowAddedCB(Object sender, WindowEventArgs args)
		{
			if (ShouldProcessWindow(args.Window) == true)
			{
				ChangeWindow(args.Window.Handle);
			}
		}
		
		private void ComponentModeChangedCB(Object sender, ComponentModeEventArgs args)
		{
			if (args.ComponentMode == ComponentMode.InAWindow && ShouldProcessWindow(args.Window) == true)
			{
				ChangeWindow(args.Window.Handle);
			}
		}
	
		private bool ShouldProcessWindow(IWindow2 window)
		{
			INode2 node = window.Node;
			
			//skip Kommunikator and VL Window
			if (node == null)
			{
				return false;
			}
			//skip some builtins
			string sysname = node.NodeInfo.Systemname;
			if (sysname == "ProjectExplorer (VVVV)" || sysname == "Finder (VVVV)" || sysname == "Inspektor (VVVV)")
			{
				return false;
			}		
			//skip InABox "Windows" .. trial & error says this is how to detect that case .. 
			if (window.IsVisible == true && node.Window == null)
			{
				return false;
			}
			//skip fullscreen (alt-enter) windows .. trial & error says this handles the case .. no beauty here
			if (GetWindowLong(window.Handle, GWL_STYLE) == -1795162112)
			{
				return false;
			}
			
			return true;
		}
		
		private void ChangeWindow(IntPtr hWnd)
		{
			//Change Window Appearance
			
			int style;
			
			style = WS_EX_WINDOWEDGE | WS_EX_APPWINDOW;
			SetWindowLong(hWnd, GWL_EXSTYLE, style);
			
			style = WS_CAPTION | WS_SYSMENU | WS_VISIBLE | WS_SIZEBOX | WS_MAXIMIZEBOX | WS_MINIMIZEBOX;			
			SetWindowLong(hWnd, GWL_STYLE, style);
			
			//Update Window Menu 
			
			IntPtr hMenu = new IntPtr(GetSystemMenu(hWnd, false));
			
			uint menuItemCount = (uint)GetMenuItemCount(hMenu);
			
			if (menuItemCount == 7) // make sure we do this only once
			{
				RemoveMenu(hMenu, menuItemCount-1, MF_BYPOSITION); // Close Command
				RemoveMenu(hMenu, menuItemCount-2, MF_BYPOSITION); // Seperator above Close
			
				DrawMenuBar(hWnd);
			}
		}
		
		//called when data for any output pin is requested
		public void Evaluate(int SpreadMax)
		{
		}
	}
}